"""

Python Desktop Server - Macros Package

Copyright (c) 2002, Georg Bauer <gb@murphy.bofh.ms>

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 
the Software, and to permit persons to whom the Software is furnished to do so, 
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""

# $Id: MacrosTool.py,v 1.71 2003/06/16 20:31:08 gb Exp $

import re
import os
import sys
import time
import stat
import calendar
import urlparse
import mimetypes
import PyDS.Tool
from PyDS.Stripogram import html2text, html2safehtml
from PyDS.strptime import strptime

# this matches any character not in the 7bit range (both plain and encoded)
highcharRE = re.compile(r'([^\x00-\x7f])')
highcharXMLRE = re.compile(r'&#(\d+);')

# this matches control characters
controlcharRE= re.compile(r'([\x00-\x1f])')

# multiple whitespace
multipleWhiteSpace = re.compile(r'(\s+)', re.DOTALL)
		
class MacrosTool(PyDS.Tool.StandardTool):

	def _initdb(self):
		try:
			self._acquire()
			self.macros = self.db.getas(
				"macros[name:S,title:S,source:S,error:S,publish:I]"
			).ordered(1)
			self._commit()
		finally: self._release()
	
	def _backup(self):
		return self._backupTable('macros', exceptcols=['error'])

	def _initopts(self):
		try:
			self._acquire()
			self.macrocache = {}
			for m in self.macros:
				self._cacheMacro(m.name)
			self._hookIntoRenderTool(self._renderIndex)
		finally: self._release()

	def _renderIndex(self):
		try:
			self._acquire()
			req = self._getDummyRequest()
			html = "<ul>"
			for macro in self.macros:
				if macro.publish and macro.error == '':
					html += '<li><b><a href="%s">%s</a>:</b> %s</li>' % (
						self.getStaticUrl(page='%s.html' % macro.name),
						macro.name,
						macro.title
					)
			html += "</ul>"
			src = req.renderPage('BaseRendering',
				body=html,
				category=''
			)
			self.storeHTML(src)
		finally: self._release()
	
	def _renderMacro(self, key, data):
		try:
			self._acquire()
			(idx, found) = self.macros.locate({'name':key})
			if found:
				macro = self.macros[idx]
				html = _('You can copy and paste the source of the following macro into your own macro edit box.')
				html += '<pre class="whiteboxsmall">'
				html += self.quoteXML(macro.source)
				html += '</pre>'
				req = self._getDummyRequest()
				src = req.renderPage('BaseRendering',
					body=html,
					category='',
					macroid=key
				)
				self.storeHTML(src, page='%s.html' % key)
		finally: self._release()
	
	def _deleteRenderedMacro(self, key, data):
		try:
			self._acquire()
			self.deleteFile(name='%s.html' % key)
		finally: self._release()

	def _rerender(self):
		try:
			self._acquire()
			for macro in self.macros:
				if macro.publish:
					self.queue.append(
						self._renderMacro,
						macro.name,
						None
					)
				else:
					self.queue.append(
						self._deleteRenderedMacro,
						macro.name,
						None
					)
		finally: self._release()
	
	def _getMacro(self, name):
		try:
			self._acquire()
			(idx, found) = self.macros.locate({'name':name})
			if found:
				macro = self.macros[idx]
				return {
					'name': macro.name,
					'title': macro.title,
					'source': macro.source,
					'publish': macro.publish
				}
			else:
				return None
		finally: self._release()
	
	def _getMacros(self):
		try:
			self._acquire()
			res = []
			for macro in self.macros:
				res.append({
					'name': macro.name,
					'title': macro.title,
					'source': macro.source,
					'publish': macro.publish,
				})
			return res
		finally: self._release()
		
	def _delMacro(self, name):
		try:
			self._acquire()
			todel = []
			idx = 0
			for macro in self.macros:
				if macro.name == name:
					todel.append(idx)
				idx += 1
			todel.sort()
			todel.reverse()
			for idx in todel:
				self.macros.delete(idx)
				if self.macrocache.has_key(name):
					del self.macrocache[name]
				self._commit()
			self.queue.append(
				self._deleteRenderedMacro,
				name,
				None
			)
		finally: self._release()
		
	def _defMacro(self, name, title, source, publish=0):
		try:
			self._acquire()
			oldpublish = 0
			(idx, found) = self.macros.locate({'name':name})
			if not(found):
				self.macros.append({'name':name, 'title':title, 'source':source, 'publish':publish})
			else:
				oldpublish = self.macros[idx].publish
				self.macros[idx].title = title
				self.macros[idx].source = source
				self.macros[idx].publish = publish
			self._commit()
			if publish:
				self.queue.append(
					self._renderMacro,
					name,
					None
				)
			else:
				if oldpublish:
					self.queue.append(
						self._deleteRenderedMacro,
						name,
						None
					)
		finally: self._release()
		self._cacheMacro(name)
		
	def _cacheMacro(self, name):
		try:
			self._acquire()
			(idx, found) = self.macros.locate({'name':name})
			if found:
				macro = self.macros[idx]
				try:
					code = compile(macro.source, macro.name, 'single')
					eval(code)
					fun = eval(macro.name)
					self.macrocache[macro.name] = lambda *args, **kw: apply(fun, (self,)+args, kw)
					macro.error = ''
				except:
					e = sys.exc_info()
					macro.error = self.error(None, e)
				self._commit()
		finally: self._release()
		
	def __getattr__(self, name):
		return self.macrocache[name]

	def _status(self):
		try:
			self._acquire()
			html = PyDS.Tool.StandardTool._status(self)
			html += '<ul>'
			html += _('<li>There are %s macros in the database</li>') % len(self.macros)
			html += _('<li>There are %s macros in the cache</li>') % len(self.macrocache.keys())
			html += _('<li>There are %s macros in the render queue</li>') % self.queue.length()
			html += '</ul>'
			return html
		finally: self._release()
	
	# this method returns the page title for rendering
	def getPageTitle(self, context):
		if context.has_key('macroid'):
			macro = self._getMacro(context['macroid'])
			return macro['name']
		else:
			return _('All Published Macros')

	# this method returns the page description for rendering
	def getPageDescription(self, context):
		if context.has_key('macroid'):
			macro = self._getMacro(context['macroid'])
			return macro['title']
		else:
			return PyDS.Tool.StandardTool.getPageDescription(self, context)

	def index_html(self, req):
		t = _('Click to add a new macro')
		html = '<p><a href="%s" title="%s">%s</a>' % (
			self.getUrl('edit_html'), t,t
		)
		html += "<ul>"
		try:
			self._acquire()
			for macro in self.macros:
					etext = ''
					if macro.error:
						etext = _('ERROR! ')
					html += '<li>%s<b><a href="%s">%s</a>:</b> %s</li>' % (
						etext,
						self.getUrl('edit_html', name=macro.name),
						macro.name,
						macro.title
					)
		finally: self._release()
		html += "</ul>"
		req.setLocalValue('title', _("All defined macros"))
		return req.renderPage('BaseTemplate', body=html)
		
	def edit_html(self, req):
		name = req.getFirstValue('name')
		title = ''
		source = ''
		error = ''
		publish = 0
		msg = ''
		if name:
			try:
				self._acquire()
				(idx, found) = self.macros.locate({'name':name})
				if found:
					m = self.macros[idx]
					title = m.title
					source = m.source
					error = m.error
					publish = m.publish
				else:
					msg = _('could not find your macro %s') % name
			finally: self._release()
		html = ''
		if msg:
			html += '<p class="errormsg">%s' % msg
		if error:
			html += error
		html += '<pre>%s</pre>' % self.numberLines(self.quoteXML(source))
		html += '<form action="%s" method="POST">' % self.getUrl('edit_redir')
		if name:
			html += '<input type="hidden" name="name" value="%s">' % name
		html += '<table border=0>'
		html += '<tr>'
		if name:
			html += '<th align="right">%s</th><td>%s</td>' % (_('Name'), name)
		html += '</tr><tr>'
		html += '<th align="right">%s</th><td><input name="title" value="%s" size=50></td>' % (
			_('Title'), title
		)
		html += '</tr><tr>'
		html += '<th align="right">%s</th><td>' % _('Publish')
		html += '<select name="publish">'
		if publish:
			html += '<option value="1" selected>%s' % _('yes')
			html += '<option value="0">%s' % _('no')
		else:
			html += '<option value="1">%s' % _('yes')
			html += '<option value="0" selected>%s' % _('no')
		html += '</select>'
		html += '</td>'
		html += '</tr><tr>'
		html += '<th align="right">%s</th><td><textarea nowrap name="source" rows=10 cols=60>%s</textarea></td>' % (
			_('Source'), self.quoteXML(source)
		)
		html += '</tr>'
		html += '<tr><td colspan=2 align="right">'
		html += '<input type=submit name="MacrosDelete" value="%s">' % _('delete')
		html += ' <input type=submit name="MacrosSubmit" value="%s">' % _('save')
		html += '</td></tr>'
		html += '</table>'
		html += '</form>'
		html += '<p align="right"><a href="%s">%s</a>' % (
			self.getUrl(),
			_('choose a different macro')
		)
		req.setLocalValue('title', name)
		return req.renderPage('BaseTemplate', body=html)
		
	def edit_redir(self, req):
		name = req.getFirstValue('name')
		if req.getFirstValue('MacrosDelete'):
			self._delMacro(name)
			return self.getUrl()
		else:
			title = req.getFirstValue('title')
			source = req.getFirstValue('source')
			publish = int(req.getFirstValue('publish'))
			if not(name):
				m = re.match(r'^def +([^\(]+)\(', source)
				if m:
					name = m.group(1)
				else:
					return self.getUrl('edit_html', errmsg=_('You must enter a function definition'))
			if title and source:
				source = source.replace('\r\n','\n')
				source = source.replace('\r', '\n')
				if source[-1:] != '\n':
					source += '\n'
				self._defMacro(name, title, source, publish)
				return self.getUrl('edit_html', name=name)
			return self.getUrl('edit_html', name=name, errmsg=_('You need to give title and source!'))
	
	# this macro formats a filesize to be read by humans
	def formatFileSize(self, num):
		fmt = '%d Byte' % num
		if num > 1000:
			fmt = '%5.2f KByte' % (num / 1000)
			num = num / 1000
			scale = ' KByte'
		if num > 1000:
			fmt = '%5.2f MByte' % (num / 1000)
			num = num / 1000
			scale = ' MByte'
		return fmt

	# This macro joins URL parts together and prevents double /
	def joinUrl(self, *parts):
		url = ''
		for p in parts:
			if p:
				if url:
					if url[-1:] == '/' and p[0] == '/':
						url += p[1:]
					else:
						url += p
				else:
					url += p
		return url
		
	# This macro returns the absolute URL for a file. pass in the relative path.
	def absoluteUrl(self, path):
		upstream = self.getToolByNamespace('upstream')
		return self.joinUrl(upstream.get('prefs', 'cloudurl'), path)
	
	# this macro numbers lines of a text and prepends line numbers
	def numberLines(self, str):
		liste = str.split('\n')
		res = ''
		num = 0
		for line in liste:
			num += 1
			res += '%4d: %s\n' % (num, line)
		return res
	
	# this macro changes all 8bit chars to &#xxx; entities. Only works
	# with latin-1, not with unicode.
	def no8bit(self, str):
		return re.sub(
			r'([\x7f-\xff])',
			lambda m: '&#%d;' % ord(m.group(1)),
			str
		)
		
	# this macro strips HTML. Used to build excerpts or other
	# strings that shouldn't have HTML. It tries to build a
	# simple text format out of HTML tags.
	def stripHTML(self, html):
		return html2text(html)
	
	# this macro strips bad HTML and only keeps good HTML
	# good HTML are those tags that are listed in valid or additional
	# usefull for excerpts that can contain HTML. Open tags
	# will be closed, so you usually get valid HTML back.
	def safeHTML(self, html, valid=('blockquote', 'ul', 'ol', 'dl', 'li', 'dt', 'dd', 'b', 'a', 'i', 'br', 'p', 'em', 'strong', 'img', 'pre', 'code'), additional=()):
		tags = {}
		for t in valid:
			tags[t] = 1
		for t in additional:
			tags[t] = 1
		return html2safehtml(html, tags.keys())
		
	# this macro removes critical characters for XML usage.
	# currently the apostrophe is not changed, as this
	# macro is used for HTML quoting, too, and HTML doesn't
	# understand apos entity tags.
	def quoteXML(self, xml, controlchars=0):
		if type(xml) == type(u''):
			xml = xml.encode('ISO-8859-1')
		xml = xml.replace('&', '&amp;')
		xml = xml.replace('<', '&lt;')
		xml = xml.replace('>', '&gt;')
		xml = xml.replace('"', '&quot;')
		xml = highcharRE.sub(lambda match: '&#%d;' % ord(match.group(1)), xml)
		if controlchars:
			xml = controlcharRE.sub(lambda match: '&#%d;' % ord(match.group(1)), xml)
		return xml
	
	# this macro unquotes a source. This is the reverse
	# to quoteXML
	def unquoteXML(self, xml):
	
		def exchar(match):
			c = int(match.group(1))
			if c >= 256:
				return '&#%d;' % c
			else:
				return chr(c)
				
		xml = highcharXMLRE.sub(exchar, xml)
		xml = xml.replace('&quot;', '"')
		xml = xml.replace('&gt;', '>')
		xml = xml.replace('&lt;', '<')
		xml = xml.replace('&amp;', '&')
		if type(xml) == type(u''):
			xml = xml.encode('ISO-8859-1')
		return xml
	
	# this macro unquotes XML stuff repeatedly, until there
	# is nothing to unquote - this can be used to unquote
	# stuff that is quoted twice
	def repeatedUnquoteXML(self, xml):
		while xml.find('&') >= 0:
			newxml = self.unquoteXML(xml)
			if newxml == xml:
				break
			xml = newxml
		return xml

	# this macro removes critical characters from an URL query.
	# currently this only quotes the ampersand in parts of the
	# query and the # before the fragment.
	def quoteURL(self, url):
		p = list(urlparse.urlparse(url))
		if p[4]:
			p[4] = p[4].replace('&', '%26')
		return urlparse.urlunparse(p).replace('#', '%23')
		
	# this macro returns a date string for a given reference time
	# (given as Float) or for the current time, if no reference time
	# is given. You can optionally add the time. This is in localtime.
	def Today(self, now=None, withtime=0):
		if now:
			now = time.localtime(float(now))
		else:
			now = time.localtime(time.time())
		if withtime:
			return time.strftime("%Y-%m-%d %H:%M:%S", now)
		else:
			return time.strftime("%Y-%m-%d", now)

	# this macro returns only the time part of a reftime
	def Time(self, now=None):
		if now:
			now = time.localtime(float(now))
		else:
			now = time.localtime(time.time())
		return time.strftime("%H:%M:%S", now)
		
	# this macro returns a standard time string in gmtime
	def TodayFull(self, now=None):
		if now:
			now = time.gmtime(float(now))
		else:
			now = time.gmtime(time.time())
		return time.strftime('%a, %d %b %Y %H:%M:%S GMT', now)
	
	# this macro parses a standard time string given by TodayFull
	def parseTodayFull(self, date):
		return calendar.timegm(strptime(date, '%a, %d %b %Y %H:%M:%S GMT'))
	
	# This macro creates the onClick attribute for a link with given parameters
	# to open a new window for this link.
	def onClick(self, id, width=515, height=480, location=0, resizable=1, scrollbars=1, status=0, toolbar=0, directories=0):
		html = "window.open (this.href, '%s', " % id
		html += "'width=%d," % width
		html += ' height=%d,' % height
		html += ' location=%d,' % location
		html += ' resizable=%d,' % resizable
		html += ' scrollbars=%d,' % scrollbars
		html += ' status=%d,' % status
		html += ' toolbar=%d,' % toolbar
		html += ' directories=%d' % directories
		html += "'); return(false);"
		return 'onclick="%s"' % html
			
	# This macro creates a permalink for a weblog posting. This is
	# a complete HTML a-tag.
	def weblogPermaLink(self, id, reftime, category='', linktext='#'):
		weblog = self.getToolByNamespace('weblog')
		upstream = self.getToolByNamespace('upstream')
		html = '<a href="%s" title="%s">%s</a>' % (
			weblog.getCloudUrlForItem(reftime, id, category),
			_("Permalink to this posting"),
			linktext
		)
		return html
	
	# This macro returns the comments script loader. This should be
	# called once for a page where you need comment functionality.
	def commentsScript(self):
		upstream = self.getToolByNamespace('upstream')
		html = '<script src="%s?u=%s&amp;c=counts" type="text/javascript"></script>' % (
			upstream.get('prefs','commentsurl'),
			upstream.get('prefs','usernum')
		)
		return html

	# This macro returns the URL for comments (used in RSS)
	def weblogCommentsUrl(self, id, reftime, category=''):
		upstream = self.getToolByNamespace('upstream')
		weblog = self.getToolByNamespace('weblog')
		url = '%s?u=%s&amp;p=%s&amp;link=%s' % (
			upstream.get('prefs','commentsurl'),
			upstream.get('prefs','usernum'),
			id,
			self.quoteURL(weblog.getCloudUrlForItem(reftime, id, category))
		)
		return url
	
	# This macro returns the comments text and link for postings
	def weblogCommentsLink(self, id, reftime, category=''):
		upstream = self.getToolByNamespace('upstream')
		src = '<a href="%s"'
		src += ' %s title="%s">%s'
		src += ' [<script type="text/javascript" '
		src += 'language="JavaScript">commentCounter(\'%s\')</script>'
		src += ']</a>'
		html = src % (
			self.weblogCommentsUrl(id, reftime, category),
			self.onClick('comments'),
			_('Click here to comment this post'),
			_('comment'),
			id
		)
		return html
		
	# This macro returns the trackback script loader. This should be
	# called once for a page where you need trackback functionality.
	def trackbackScript(self):
		upstream = self.getToolByNamespace('upstream')
		html = '<script src="%s?u=%s&amp;c=counts" type="text/javascript"></script>' % (
			upstream.get('prefs','trackbackurl'),
			upstream.get('prefs','usernum')
		)
		return html

	# This macro returns the URL for trackbacks (used in RSS)
	def weblogTrackbackUrl(self, id, reftime, category=''):
		upstream = self.getToolByNamespace('upstream')
		weblog = self.getToolByNamespace('weblog')
		url = '%s?u=%s&amp;p=%s' % (
			upstream.get('prefs','trackbackurl'),
			upstream.get('prefs','usernum'),
			id
		)
		return url
	
	# This macro returns the trackback text and link for postings
	def weblogTrackbackLink(self, id, reftime, category=''):
		upstream = self.getToolByNamespace('upstream')
		src = '<a href="%s"'
		src += ' %s title="%s">%s'
		src += ' [<script type="text/javascript" '
		src += 'language="JavaScript">trackbackCounter(\'%s\')</script>'
		src += ']</a>'
		html = src % (
			self.weblogTrackbackUrl(id, reftime, category),
			self.onClick('trackbacks'),
			_('Click here to see the trackbacks for this post'),
			_('trackback'),
			id
		)
		return html
		
	# This macro returns the webbug script for referrer tracking
	def webbug(self):
		upstream = self.getToolByNamespace('upstream')
		html = '<script language="JavaScript" type="text/javascript"><!--'
		html += '\nvar imageUrl = "%s";' % upstream.get('prefs','webbugurl')
		html += '\nvar imageTag = "<img src=\\"" + imageUrl + "?group=default'
		html += '&usernum=%s' % upstream.get('prefs','usernum')
		html += '&referer=" + escape (document.referrer) + "\\" height=\\"1\\" width=\\"1\\">";'
		html += '\ndocument.write (imageTag);'
		html += '\n//--></script><p>'
		return html
	
	# this macro returns an image tag for a blind (transparent) image.
	# Usually used to span tables over defined areas.
	def blind(self, sizex, sizey):
		return '<img src="%s" width=%d height=%d border=0>' % (
			self.absoluteUrl('/images/blind.gif'), sizex, sizey
		)
		
	# This macro returns an image tag for a local image. pass in the relative path.
	# If you give an url, a link is created automatically.
	def imageTag(self, src, alt=None, width=None, height=None, border=0, align=None, url=None):
		html = '<img src="%s"' % self.absoluteUrl(src)
		if _flet.desktop and len(src) > 8 and src[:8] == '/images/':
			html = '<img src="%s"' % src
		if border != None:
			html += ' border=%d' % border
		if width:
			html += ' width=%d' % width
		if height:
			html += ' height=%d' % width
		if alt != None:
			html += ' alt="%s"' % alt
		if align:
			html += ' align="%s"' % align
		html += '>'
		if url:
			html = '<a href="%s" title="%s">%s</a>' % (
				url,
				alt,
				html
			)
		return html

	# This macro constructs a simple link with given target (default
	# _blank).  You can pass in the named parameters for onclick, if you
	# do, a suiting onclick attribute will be generated. In that case the
	# target must be different from "_blank"! An empty target removes
	# the target= attribute from the link.
	def linkTag(self, url, title='', target="_blank", **kw):
		if not(title):
			title = url
		if kw != {} and target!='_blank':
			html = '<a target=%s" href="%s" %s title="%s">%s</a>' % (
				target,
				url,
				apply(self.onclick, [target], kw),
				title,
				title
			)
		elif target:
			html = '<a target="%s" href="%s" title="%s">%s</a>' % (
				target,
				url,
				title,
				title
			)
		else:
			html = '<a href="%s" title="%s">%s</a>' % (
				url,
				title,
				title
			)
		return html
	
	# this macro links to a weblog posting
	def linkPost(self, id):
		weblog = self.getToolByNamespace('weblog')
		post = weblog.getPost(id)
		title = ''
		if post:
			title = post['title']
			category = ''
			if not(post['onhome']):
				categories = post['categories']
				if len(categories):
					category = categories[0]
				else:
					post = {}
		if post:
			url = weblog.getCloudUrlForItem(
				post['pubtime'],
				id,
				category=category
			)
			if not(title):
				title = url
			html = '<a href="%s" title="%s">%s</a>' % (
				url,
				title,
				title
			)
			return html
		else:
			return ''
		
	# this macro constructs a link to a macro
	def linkMacro(self, macroname):
		macro = self._getMacro(macroname)
		if macro and macro['publish']:
			html = '<a href="%s" title="%s">%s</a>' % (
				self.getStaticUrl(page='%s.html' % macroname),
				macro['title'],
				macro['title']
			)
			return html
		else:
			return ''

	# this macro constructs a link to a nugget
	def linkNugget(self, name):
		nuggets = self.getToolByNamespace('nuggets')
		nugget = nuggets._getNugget(name)
		if nugget and nugget['publish']:
			html = '<a href="%s" title="%s">%s</a>' % (
				nuggets.getStaticUrl(page='%s.html' % name),
				nugget['title'],
				nugget['title']
			)
			return html
		else:
			return ''

	# this macro constructs a link to a picture
	def linkPicture(self, id, format=None):
		pictures = self.getToolByNamespace('pictures')
		pic = pictures.getPicture(id)
		if pic and pic['publish']:
			if format:
				format = pictures.getFormat(format)
			if not(format):
				format = pictures.getFormatList()[0]
			html = '<a href="%s" title="%s">%s</a>' % (
				pictures.getStaticUrl(page='%s.html' % id, category=format['name']),
				pic['title'],
				pic['title']
			)
			return html
		else:
			return ''

	# this macro constructs a link to a story
	def linkStory(self, storyid, withchange=0):
		stories = self.getToolByNamespace('stories')
		story = stories.getStory(storyid)
		if story:
			html = '<a href="%s" title="%s">%s</a>' % (
				stories.getStoryUrl(storyid),
				story['title'],
				story['title']
			)
			if withchange:
				html += ' <span class="smallfont">(%s %s)</span>' % (
					_('last change'),
					self.Today(now=story['pubtime'], withtime=1)
				)
			return html
		else:
			return ''
			
	# this macro returns the copyright message for a given year-range
	def copyright(self, fromyear, toyear):
		prefs = self.getPreferencesTool()
		if fromyear > toyear:
			x = toyear
			toyear = fromyear
			fromyear = x
		if fromyear != toyear:
			return '&copy; %s-%s, %s' % (
				fromyear,
				toyear,
				prefs.get('private', 'name')
			)
		else:
			return '&copy; %d, %s' % (
				fromyear,
				prefs.get('private', 'name')
			)

	# this macro returns a link to the mailto function
	def mailtoLink(self):
		upstream = self.getToolByNamespace('upstream')
		html = '<a href="%s?usernum=%s" title="%s">%s</a>' % (
			upstream.get('prefs','mailtourl'),
			upstream.get('prefs','usernum'),
			_('Mail the weblog author'),
			self.imageTag('/images/mailto.gif', alt=_('Letterimage'))
		)
		return html

	# this macro returns a link to the RSS feed. The current tool and
	# the category must be passed in as parameter. If the tool doesn't
	# have RSS availability, nothing is returned. The title can be
	# overridden
	def rssLink(self, tool, category, title=None):
		if type(tool) == type(''):
			tool = self.getToolByNamespace(tool)
		if not(title):
			title = _('Syndicate this feed with XML')
		if tool.rss:
			html = '<a href="%s" title="%s">%s</a>' % (
				tool.getStaticUrl(page='rss.xml',category=category),
				title,
				self.imageTag('/images/miniXmlButton.gif', alt=_('XML-Image'))
			)
			return html
		else:
			return ''
	
	# this macro just copies a file to the gems folder and so
	# upstreams it to the cloud. It returns the url of the file
	# in the cloud and the relative name below the gems folder.
	# http:// urls are supported, too! If savecopy is true,
	# always a local copy is saved. Otherwise only local files are
	# copied to the gems folder!
	def downloadUrl(self, localname, remotename=None, savecopy=0):
		isLocal = 1
		if len(localname) > 7 and localname[:7] == 'http://':
			downstream = self.getToolByNamespace('downstream')
			f = downstream.download(localname)
			isLocal = 0
		else:
			f = open(localname)
		content = f.read()
		f.close()
		(path, base) = os.path.split(localname)
		if remotename:
			base = remotename
		if savecopy or isLocal:
			fname = os.path.join(_PyDS.ROOTDIR, 'gems', base)
			(lpath, lbase) = os.path.split(fname)
			if not(os.path.isdir(lpath)):
				os.makedirs(lpath)
			oldcontent = None
			if os.path.exists(fname):
				f = open(fname)
				oldcontent = f.read()
				f.close()
			if content != oldcontent:
				f = open(fname, 'w')
				f.write(content)
				f.close()
			return (self.absoluteUrl('/gems/%s' % base), base)
		else:
			return (localname, base)

	# this macro returns a box with the files infos and a download
	# link beneath that. Files are automatically copied to the gems
	# folder and so published to the cloud. If savecopy is true,
	# always a local copy is saved. Otherwise only copies for local
	# files are saved, remote files return their remote url.
	def downloadLinkBox(self, localname, remotename=None, savecopy=0):
		downstream = self.getToolByNamespace('downstream')
		(url, name) = self.downloadUrl(localname, remotename, savecopy=savecopy)
		html = '<p class="whiteboxsmall">'
		mtype = mimetypes.guess_type(localname)[0]
		if not(mtype):
			mtype = 'unknown/unknown'
		html += '<b>%s:</b> ' % name
		fields = None
		lmod = None
		if len(localname) > 7 and localname[:7] == 'http://':
			fields = os.stat(downstream.getCachePath(localname))
			lmod = fields[stat.ST_MTIME]
			lmodhttp = downstream.getModificationTime(localname)
			if lmodhttp:
				lmod = lmodhttp
		else:
			fields = os.stat(localname)
			lmod = self.TodayFull(now=fields[stat.ST_MTIME])
		html += _('%s, last change: %s, type: %s') % (
			self.formatFileSize(fields[stat.ST_SIZE]),
			lmod, mtype
		)
		html += '<p class="smallfont" align="right">'
		html += '<a href="%s">%s</a>' % (
			url,
			_('click here to download file')
		)
		return html
	
	# this macro returns a box with the files content and a download
	# link beneath that. Files are automatically copied to the gems
	# folder and so published to the cloud. savecopy is the same as
	# with downloadLinkBox (if false, no local copy is saved for remote
	# urls)
	def downloadSourceBox(self, localname, remotename=None, savecopy=0):
		downstream = self.getToolByNamespace('downstream')
		(url, name) = self.downloadUrl(localname, remotename, savecopy=savecopy)
		html = '<pre class="whiteboxsmall">'
		f = None
		if len(localname) > 7 and localname[:7] == 'http://':
			f = open(downstream.getCachePath(localname))
		else:
			f = open(localname)
		content = f.read()
		f.close()
		html += self.quoteXML(content)
		html += '</pre>'
		html += '<p class="smallfont" align="right">'
		html += '<a href="%s">%s</a>' % (
			url,
			_('click here to download file')
		)
		return html

__desc__ = _("This is a tool that manages macros for text rendering.")

PyDS.Tool.registerTool(MacrosTool('macros', _('Macros'), __desc__, 650))

