"""

Python Desktop Server - Shortcut Editor

Copyright (c) 2002, Garth T Kidd <garth@deadlybloodyserious.com>

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 
the Software, and to permit persons to whom the Software is furnished to do so, 
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""

# $Id: MeshTool.py,v 1.3 2003/06/15 16:48:00 gb Exp $

import PyDS.Tool
import PyDS.StructuredText
import metakit
import re
import docutils.transforms
import docutils.nodes
import urllib

class TargetVisitor(docutils.nodes.SparseNodeVisitor):
	"""`Visitor` to find targets, as per Georg Bauer's original code."""

	def __init__(self, document):
		docutils.nodes.SparseNodeVisitor.__init__(self, document)
		self.targetids = {}
	
	def visit_target(self, node):
		if node.has_key('name') and node.has_key('id'):
			self.targetids[node['name']] = node['id']

_toolsWithStrayLinkResolvers = None

def toolsWithStrayLinkResolvers(): 
	global _toolsWithStrayLinkResolvers
	tools = _toolsWithStrayLinkResolvers
	if tools is not None: 
		return tools
	tools = [tool for tool in PyDS.Tool.getTools()
	         if hasattr(tool, 'getCloudUrlMatchingStrayLink')]
	_toolsWithStrayLinkResolvers = tools
	return tools

class StrayLinkResolver(docutils.nodes.SparseNodeVisitor):
	"""`Visitor` to resolve stray links.
	
	Intended to facilitate links to other weblog postings and anything
	else another tool might come up with. Say, Wiki postings."""

	def __init__(self, document, targetVisitor):
		"""Initialise the `StrayLinkResolver` with a given `TargetVisitor`
		instance. The latter should have already been walked over the
		document."""
		docutils.nodes.SparseNodeVisitor.__init__(self, document)
		self.targetVisitor = targetVisitor
		self.foundTools = toolsWithStrayLinkResolvers()
		
	def resolveStrayLink(self, stray): 
		"""Resolve a stray by consulting all tools capable of doing so."""
		foundTools = self.foundTools
		if [t for t in foundTools if t.name == _flet.current.name]:
			resolvers = [_flet.current] 
		else: 
			resolvers = []
		resolvers.extend([t for t in foundTools if t.name != _flet.current.name])
		for tool in resolvers: 
			match = tool.getCloudUrlMatchingStrayLink(stray)
			if match: 
				return match

	def visit_reference(self, node):
		"""Fix unresolvable references."""
		if node.resolved and node.has_key('refuri'):
			refuri = node['refuri']
			(method, rest) = urllib.splittype(refuri)
			if method is None: 
				match = self.resolveStrayLink(refuri)
				if match: 
					cssclass, uri = match
					node['class'] = cssclass
					node['refuri'] = uri
			return
		if node.resolved \
			or not node.hasattr('refname') \
			or self.targetVisitor.targetids.has_key(node['refname']):
			return
		match = self.resolveStrayLink(node['refname'])
		if match:
			cssclass, uri = match
			node.resolved = 1
			node['class'] = cssclass
			node['refuri'] = uri
			del node['refname']

class StrayLinkTransform(docutils.transforms.Transform):
	"""`Transform` to resolve stray links across a document."""
	def transform(self):
		if toolsWithStrayLinkResolvers(): 
			targetVisitor = TargetVisitor(self.document)
			self.document.walk(targetVisitor)
			self.document.walk(StrayLinkResolver(self.document, targetVisitor))

class MeshTool(PyDS.Tool.StandardTool):

	def _initdb(self):
		try:
			self._acquire()
			if 0:
				self.targetCache = self.db.getas(
					"[tool:S,uniqueID:S,field:S,content:S]"
				).ordered(1)
				self.prefs = self.db.getas(
					"prefs[slackness:I]"
				)
				if len(self.prefs) == 0:
					self.prefs.append(slackness=1)
					self._commit()
		finally: self._release()
	
	def _backup(self):
		return self._backupTable('prefs')
		
	def _status(self):
		try:
			self._acquire()
			html = PyDS.Tool.StandardTool._status(self)
			html += _('<p>There are %d tools registered for stray link resolving.') % len(toolsWithStrayLinkResolvers())
			return html
		finally: self._release()

	def index_html(self, req):
		html = ''
		tools = toolsWithStrayLinkResolvers()
		if tools: 
			html += '<ul>\n%s</ul>' % ''.join(
					['<li><a href="%s">%s</a></li>\n' % (
						t.getUrl(), t.title )
					 for t in tools]
				)
		req.setLocalValue('title', _('Stray Link resolving Tools'))
		return req.renderPage('BaseTemplate', body=html)
		
__desc__ = _("This tool allows to configure stray link resolution")

PyDS.Tool.registerTool(MeshTool('mesh', _('Mesh'), __desc__, 910))
PyDS.StructuredText.registerTransform(1, StrayLinkTransform)

