"""

Python Desktop Server - Tool base class

Copyright (c) 2002, Georg Bauer <gb@murphy.bofh.ms>

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 
the Software, and to permit persons to whom the Software is furnished to do so, 
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""

# $Id: Tool.py,v 1.133 2003/06/12 15:35:13 gb Exp $

import string
import re
import os
import shutil
import time
import cgi
import sys
import traceback
import urllib
import threading
import sgmllib

import metakit

from medusa import http_server
from medusa import filesys

from Cheetah.Template import Template

import PyDS.StructuredText
import PyDS

version = 'Python Desktop Server %s (http://pyds.muensterland.org/)' % PyDS.__version__

# calendar functions
import calendar
calendar.setfirstweekday(calendar.MONDAY)
wdnames = _('Mo Tu We Th Fr Sa Su').split()
mlnames = _('January February March April May June July August September October November December').split()
msnames = _('Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec').split()

# fetch the fastest StringIO available
try:
	import cStringIO
	StringIO = cStringIO
except:
	import StringIO

# regexp for sources to include in trace
tracere = None

# trace functions
def tracefun(frame, event, arg):
	from inspect import getframeinfo
	(fil, lno, fun, src, res) = getframeinfo(frame)
	if tracere and not(tracere.match(fil)): return
	sys.stderr.write("# %s:%d\n" % (fil, lno))
	if src:
		for l in src:
			sys.stderr.write(l)
	return traceloc

def traceloc(frame, event, arg):
	from inspect import getframeinfo
	(fil, lno, fun, src, res) = getframeinfo(frame)
	if tracere and not(tracere.match(fil)): return
	if src:
		for l in src:
			sys.stderr.write(l)
	if event == 'return':
		sys.stderr.write("# return %s: %s" % (fun, arg))
	return None

# this method starts tracing (logs to error.log)
def startTrace(pattern=None):
	global tracere
	if pattern: pattern = re.compile(pattern)
	tracere = pattern
	sys.settrace(tracefun)

# this method stops tracing
def stopTrace():
	sys.settrace(None)

# this method writs something to the error log
def debugOutput(msg):
	apply(sys.stderr.write, ("# "+msg+"\n",))
	
# look for a template and copy it if necessary
def getTemplate(tmpl):
	tfile = os.path.join(_PyDS.TMPLDIR, '%s.tmpl' % tmpl)
	if not(os.path.exists(tfile)):
		tfilesrc = os.path.join(_PyDS.SHAREDDIR, '%s.tmpl' % tmpl)
		if os.path.exists(tfilesrc):
			shutil.copyfile(tfilesrc, tfile)
		else:
			tfile = os.path.join(_PyDS.TMPLDIR, 'Default.tmpl')
			if not(os.path.exists(tfile)):
				tfilesrc = os.path.join(_PyDS.SHAREDDIR, 'Default.tmpl')
				shutil.copyfile(tfilesrc, tfile)
	return tfile

# Registration of installed tools and their namespaces
_ToolNamespace = {}

# Overrides from config file for tool registration
_ToolOverrides = {}
_ToolDefaults = {}
try:
	import PyDSTools
	for k in PyDSTools.__dict__:
		if k.find('_') < 0:
			_ToolOverrides[k] = getattr(PyDSTools, k)
except: pass

# register a tool into the registry
def registerTool(tool):
	_ToolDefaults[tool.name] = tool.level
	if _ToolOverrides.has_key(tool.name):
		tool.level = _ToolOverrides[tool.name]
	_ToolNamespace[tool.name] = tool

# get a list of tools in the order by level and title
def getTools():
	def cmptool(a,b):
		if a.level == b.level:
			return cmp(a.title, b.title)
		else:
			return cmp(a.level, b.level)

	liste = []
	for k in _ToolNamespace.keys():
		liste.append(_ToolNamespace[k])
	liste.sort(cmptool)
	return liste

# get a list of handlers (tools in reverse) for installation in http_server
# Tools with level >= 2000 are not loaded as handlers!
def getHandlers():
	liste = filter(lambda a: (a.level < 2000), getTools())
	liste.reverse()
	return liste

# get all tools with level < 800 for the top menubar
def getFirstToolbar():
	liste = filter(lambda a: (a.level < 800), getTools())
	return liste

# get all tools with level 8XX for the first block of tools on the right
def getSecondToolbar():
	liste = filter(lambda a: (a.level >= 800) and (a.level < 900),
		getTools())
	return liste

# get all tools with level 9XX for the second block of tools on the right
def getThirdToolbar():
	liste = filter(lambda a: (a.level >= 900) and (a.level < 1000), getTools())
	return liste

# toolfunction to install a template that isn't available - this is to transport
# default templates within tools.
def spillOutTemplate(template, src):
	tname = os.path.join(_PyDS.TEMPLATEDIR, "%s.tmpl" % template)
	if not(os.path.exists(tname)):
		tf = open(tname, "w")
		tf.write(src)
		tf.close()

# functions to make log entries without needing the tool
def log(fmt, *args):
	_ToolNamespace['events']._log('server', fmt % args)

def logError(fmt, *args):
	_ToolNamespace['events']._log('server', fmt % args, error=1)

def logVerbose(fmt, *args):
	_ToolNamespace['events']._logVerbose('server', fmt % args)

# check wether a req object is a call to a redirection
redirre = re.compile('_redir$')
def isRedir(req):
	if redirre.match(req.path):
		return 1
	else:
		return 0

# a special class only to collect all toolfunctions only one instance will
# be created and carry registered tool functions (tool.* namespace in
# templates) as attributes
class _ToolFunctionsSingleton:
	pass

_ToolFunctions = _ToolFunctionsSingleton()

# register tool functions for templates
_ToolFunctions.getFirstToolbar = getFirstToolbar
_ToolFunctions.getSecondToolbar = getSecondToolbar
_ToolFunctions.getThirdToolbar = getThirdToolbar
_ToolFunctions.getTemplate = getTemplate
_ToolFunctions.version = version

# this class is a parser for backup data. It uses sgmllib because it
# is much smaller and lightweight than expat and much more tolerant to
# problematic input
class BackupParser(sgmllib.SGMLParser):

	def reset(self):
		sgmllib.SGMLParser.reset(self)
		self._tablename = ''
		self._toolname = ''
		self._current = ''
		self._items = []
		self._itemstack = []
		self._tagstack = []
		self._inItem = 0
	
	def _push(self, tag):
		self._itemstack.append(self._item)
		self._tagstack.append(tag)
		self._item = {}
	
	def _pop(self):
		olditem = self._item
		self._item = self._itemstack.pop()
		self._tagstack.pop()
		return olditem
	
	def start_table(self, attrs):
		if self._inItem:
			self.unknown_starttag('table', attrs)
		else:
			for (a, v) in attrs:
				if a == 'name':
					self._tablename = v
				elif a == 'tool':
					self._toolname = v
	
	def start_item(self, attrs):
		if self._inItem:
			self.unknown_starttag('item', attrs)
		else:
			self._item = {}
		self._inItem += 1
	
	def end_item(self):
		self._inItem -= 1
		if self._inItem:
			self.unknown_endtag('item')
		else:
			self._items.append(self._item)
	
	def unknown_starttag(self, tag, attrs):
		self._itemtype = 'str'
		for (a, v) in attrs:
			if a == 'type':
				self._itemtype = v
				if v == 'PyView':
					self._item[tag] = []
					self._push(tag)
		self._current = ''
	
	def unknown_endtag(self, tag):
		if len(self._tagstack) and tag == self._tagstack[-1]:
			newitem = self._pop()
			self._item[tag].append(newitem)
			self._itemtype = 'PyView'
		else:
			if self._itemtype == 'int':
				self._item[tag] = int(self._current)
			elif self._itemtype == 'float':
				self._item[tag] = float(self._current)
			elif self._itemtype == 'str':
				self._item[tag] = self._current
	
	def handle_data(self, data):
		self._current += data


# this class implements a command queue for a tool. A command queue handles
# commands that might stump out other commands and are unique (identical
# commands are not executed twice in one run). A command consists of the
# command name, a key for the command to recognize identical commands and
# some data. key and data are whatever you like, but the key must be
# hashable, as it is used as a key to a hash.
class CommandQueue:

	def __init__(self):
		self.hash = {}
		self.list = []
		self.lock = threading.RLock()
	
	def __len__(self):
		return self.length()
	
	def append(self, cmd, key, data):
		try:
			self.lock.acquire()
			if self.hash.has_key(key):
				idx = self.hash[key]
				el = self.list[idx]
				el[0] = cmd
				el[2] = data
			else:
				self.list.append([cmd, key, data])
				self.hash[key] = len(self.list)-1
		finally: self.lock.release()
	
	def length(self):
		try:
			self.lock.acquire()
			return len(self.list)
		finally: self.lock.release()

	def shift(self):
		try:
			self.lock.acquire()
			if len(self.list):
				first = self.list[0]
				del self.list[0]
				del self.hash[first[1]]
				for k in self.hash.keys():
					self.hash[k] -= 1
				return first
			else:
				return None
		finally: self.lock.release()


# this class is my own version of a Request, built out of the medusa request.
# this class carries several tool functionality needed for templates. If you
# give base and flat, those parameters will be passed on to getStaticUrl
# of the relevant tools. This enables you to bind a request to a different
# storage with different structure.
class _Request:

	def __init__(self, tool, request=None, input_data=None, now=None, base=None, flat=0):
		self.request = request
		self.tool = tool
		self.shortcuts = tool.getToolByNamespace('shortcuts')
		self.prefs = tool.getPreferencesTool()
		self.dummy = 0
		self.url = tool.url
		self.base = base
		self.flat = flat
		if request:
			[path, params, query, fragment] = request.split_uri()
			self.path = path
			if path and string.find(path, '%') >= 0:
				self.path = http_server.unquote(path)
			self.params = params
			if params and string.find(params, '%') >= 0:
				self.params = http_server.unquote(params)
			if self.params and self.params[0] == ';':
				self.params = self.params[1:]
			if query:
				if query[0] == '?':
					self.query = cgi.parse_qs(query[1:])
				else:
					self.query = cgi.parse_qs(query)
			else:
				self.query = {}
		else:
			self.dummy = 1
			self.query = {}
			self.path = '/%s/' % tool.name
		if input_data:
			self.form = cgi.parse_qs(input_data.read())
		else:
			self.form = {}
		self.local = {}
		if request:
			self.fragment = fragment
			if fragment and string.find(fragment, '%') >= 0:
				self.fragment = http_server.unquote(fragment)
		else:
			self.fragment = ''
		(year, month, day, hour, minute, second) = time.localtime(time.time())[:6]
		self.thisyear = year
		self.thismonth = month
		self.thisday = day
		if now:
			(year, month, day, hour, minute, second) = time.localtime(now)[:6]
		self.tyear = year
		self.tmonth = month
		self.tday = day
		self.dateGiven = 0
		self.idGiven = 0
		self.dayGiven = 0
		try: f = map(lambda s: int(s), self.params.split('-'))
		except: f = []
		if len(f) >= 1:
			year = f[0]
			self.dateGiven = 1
		if len(f) >= 2:
			month = f[1]
			self.dateGiven = 1
		if len(f) >= 3:
			day = f[2]
			self.dateGiven = 1
			self.dayGiven = 1
		self.number = 0
		if len(f) >= 4:
			self.number = f[3]
			self.idGiven = 0
		self.year = year
		self.month = month
		self.day = day
		self.time = '%02d:%02d:%02d' % (hour, minute, second)
		
	def getAllValues(self, name, default=[]):
		if self.local.has_key(name):
			val = self.local[name]
		elif self.query.has_key(name):
			val = self.query[name]
		elif self.form.has_key(name):
			val = self.form[name]
		else:
			return default
		if type(val) != type([]):
			val = [val]
		# normalize line endings
		val = map(lambda x: x.replace('\x0d\x0a', '\n'), val)
		val = map(lambda x: x.replace('\r', '\n'), val)
		return val
		
	def getAllValuesQuoted(self, name, default=[]):
		l = self.getAllValues(name, default)
		r = []
		for e in l:
			e = e.replace('&', '&amp;')
			e = e.replace('<', '&lt;')
			e = e.replace('>', '&gt;')
			r.append(e)
		return r

	def setLocalValue(self, name, value):
		self.local[name] = value
	
	def getFirstValue(self, name, default=''):
		if self.local.has_key(name):
			val = self.local[name]
		elif self.query.has_key(name):
			val = self.query[name]
		elif self.form.has_key(name):
			val = self.form[name]
		else:
			return default
		if type(val) == type([]):
			val = val[0]
		# normalize line endings
		val = val.replace('\x0d\x0a', '\n')
		val = val.replace('\r', '\n')
		return val
		
	def getFirstValueQuoted(self, name, default=''):
		val = self.getFirstValue(name, default)
		val = val.replace('&', '&amp;')
		val = val.replace('<', '&lt;')
		val = val.replace('>', '&gt;')
		return val
	
	def getCloudTitle(self, context={}):
		title = self.getFirstValue('title')
		if not(title):
			title = self.tool.getPageTitle(context)
		return title
	
	def getDesktopTitle(self):
		prefs = self.tool.getPreferencesTool()
		title = self.getFirstValue('title')
		if not(title):
			title = prefs.get('private', 'title')
		if not(title):
			title = _('Python Desktop Server')
		if self.dateGiven:
			if self.dayGiven:
				return '%s %s: %s %s.%s.%s' % (
					_PyDS.titletag,
					PyDS.__version__,
					title,
					self.day,
					self.month,
					self.year
				)
			else:
				return '%s %s: %s %s/%s' % (
					_PyDS.titletag,
					PyDS.__version__,
					title,
					self.month,
					self.year
				)
		else:
			return '%s %s: %s' % (_PyDS.titletag, PyDS.__version__, title)

	def getUrl(self, page='index_html', tool=None, fragment=None, **kw):
		if tool:
			tool = self.tool.getToolByNamespace(tool)
		else:
			tool = self.tool
		url = '/%s/%s' % (tool.name, page)
		if self.params:
			url += ';%s' % self.params
		l = []
		for k in kw.keys():
			kv = urllib.quote_plus(kw[k])
			l.append('%s=%s' % (k, kv))
		if len(l):
			url += '?' + string.join(l, '&')
		if fragment:
			url += '#' + fragment
		return url
		
	def getStaticUrl(self, page=None, tool=None, year=None, month=None, day=None, category=''):
		if tool:
			tool = self.tool.getToolByNamespace(tool)
		else:
			tool = self.tool
		return tool.getStaticUrl(page=page, year=year, month=month, day=day, category=category, base=self.base, flat=self.flat)

	def getNamespace(self, context={}):
		return {
			'title': self.tool.title,
			'name': self.tool.name,
			'desc': self.tool.desc,
			'url': self.url,
			'current': self.tool,
			'request': self,
			'tools': _ToolFunctions,
			'prefs': _ToolNamespace['preferences'],
			'errmsg': self.getFirstValue('errmsg'),
			'infomsg': self.getFirstValue('infomsg'),
			'withcalendar': self.tool.showCalendar,
			'context': context,
			'_': _
		}

	def renderTextWithTraceback(self, tmpl, structured=0, **kw):
		try:
			html = apply(self.renderText, [tmpl, structured], kw)
			return html
		except:
			(e, d, tb) = sys.exc_info()
			msg = '<table border=0>'
			msg += '<tr><td colspan=4>'
			msg += _('<b>Exception %s: %s</b>') % (e, d)
			msg += '</td></tr>'
			for row in traceback.extract_tb(tb):
				msg += '<tr>'
				(file, line, function, source) = row
				msg += '<td>%s</td>' % file
				msg += '<td>%s</td>' % line
				msg += '<td>%s</td>' % function
				msg += '<td>%s</td>' % source
				msg += '</tr>'
			msg += '</table>'
			return msg

	def renderText(self, tmpl, structured=0, section_level=1, **kw):
		if structured:
			suppress_paragraphs = 0
			if self.prefs.get('private', 'suppara'):
				suppress_paragraphs = 1
			tmpl = PyDS.StructuredText.renderText(tmpl,
				section_level=section_level,
				suppress_paragraphs=suppress_paragraphs
			)
			tmpl = self.shortcuts.processShortcuts(tmpl, quotedelimiter='&quot;')
		else:
			tmpl = self.shortcuts.processShortcuts(tmpl)
		kw2 = self.getNamespace(kw)
		t = Template(tmpl, searchList = [kw, kw2, _ToolNamespace])
		return str(t)

	def renderPage(self, tmpl, **kw):
		kw2 = self.getNamespace(kw)
		tfile = getTemplate(tmpl)
		t = Template(file=tfile, searchList = [kw, kw2, _ToolNamespace])
		return str(t)

	def renderStaticCalendar(self, category=''):
		caltool = self.tool.getToolByNamespace('calendar')
		return caltool.render(
			self.tool,
			year=self.year,
			month=self.month,
			day=self.day,
			category=category
		)

	def renderCalendar(self, category=None):
		def getDynamicUrl(year=None, month=None, day=None, category=None):
			if day:
				return '%s;%04d-%02d-%02d' % (self.path, year, month, day)
			else:
				return '%s;%04d-%02d' % (self.path, year, month)
		
		caltool = self.tool.getToolByNamespace('calendar')
		urlcallback = getDynamicUrl
		if hasattr(self.tool, 'getCalendarUrl'):
			urlcallback = self.tool.getCalendarUrl
		return caltool.render(
			self.tool,
			withtoday=1,
			urlcallback=urlcallback,
			year=self.year,
			month=self.month,
			day=self.day,
			category=category
		)


# just a standard collector for POST data
class Collector:

	def __init__(self, handler, length, request):
		self.handler = handler
		self.request = request
		self.request.collector = self
		self.request.channel.set_terminator(length)
		self.buffer = StringIO.StringIO()

	def collect_incoming_data(self, data):
		self.buffer.write(data)

	def found_terminator(self):
		self.buffer.seek(0)
		self.request.collector = None
		self.request.channel.set_terminator('\r\n\r\n')
		self.handler.continue_request(self.request, self.buffer)


# a tool without functionality. Only external links to the cloud
class LinkTool:

	def __init__(self, name, title=None, desc=None, url=None, level=800):
		self.name = name
		self.title = title
		if not(title):
			self.title = name
		self.desc = desc
		if not(self.desc):
			self.desc = _("This is a link-only tool")
		self.url = url
		self.level = level
		self.hasAPI = 0

	def nugget(self, url=None):
		if not(url):
			url = self.url
		html = '<a href="%s">%s</a>: %s' % (url, self.title, self.desc)
		return html
	
	def needsStuff(self, req):
		return 0

# this class encapsulates hooks into a tool
class StandardHook:

	def __init__(self, toolname, mode, tag, title, func, activeFunc):
		self.title = title
		self._func = func
		self._activeFunc = activeFunc
		self._tool = toolname
		self._mode = mode
		self._tag = tag
	
	def getId(self):
		return '%s.%s.%s' % (self._tool, self._mode, self._tag)
	
	def isActive(self, *args, **kw):
		if self._activeFunc:
			return apply(self._activeFunc, args, kw)
		else:
			return 1
	
	def __call__(self, *args, **kw):
		return apply(self._func, args, kw)
		
		
# a standard tool with functionality
class StandardTool:

	def __init__(self, name, title=None, desc=None, level=400):
		self.name = name
		self.title = title
		if not(title):
			self.title = self.name
		self.desc = desc
		if not(self.desc):
			self.desc = _("This is a standard tool")
		self.url = "/%s/" % self.name
		self.level = level
		self.regex = re.compile(r'^/%s/([^/]+)$' % name)
		self.db = metakit.storage(
			os.path.join(_PyDS.VARDIR, '%s.data' % name), 1)
		self.rss = 0
		self.hasAPI = 1
		self.lock = threading.RLock()
		self.queue = CommandQueue()
		self.hooks = {}
		self.showCalendar = 1
		self.useAuthentication = 1
		self.needAdmin = 1

	def _restoreGetTable(self, tablename):
		return getattr(self, tablename, None)
	
	def _restoreTable(self, table, items, purge=0):
		try:
			self._acquire()
			if purge:
				l = range(len(table))
				l.reverse()
				for idx in l:
					table.delete(idx)
			for item in items:
				views = []
				for k in item.keys():
					if type(item[k]) == type([]):
						views.append((k, item[k]))
						item[k] = metakit.view()
				idx = table.append(item)
				for (col, view) in views:
					c = getattr(table[idx], col)
					for el in view:
						c.append(el)
			self._commit()
		finally: self._release()

	def _backupTable(self, tablename, tableobj=None, exceptcols=[], sortcols=[], groupfn=None, selectfn=None):
		def _storeBackup(xml, last, lastcols, cloud):
			out = '<table tool="%s" name="%s"' % (self.name, tablename)
			for (cn, cv) in lastcols:
				out += ' %s="%s"' % (cn, cv)
			out += '>\n'
			out += xml
			out += '</table>\n'
			return self.storeXML(
				out,
				name=('%s.xml' % last),
				category=self.name,
				toolname='backup',
				cloud=cloud
			)

		def _valType(val):
			typ = str(type(val))
			typ = typ.split(' ')[1]
			typ = typ.replace('>','')
			typ = typ.replace("'",'')
			return typ

		try:
			self._acquire()
			excols = {}
			for ec in exceptcols:
				excols[ec] = 1
			macros = self.getToolByNamespace('macros')
			prefs = self.getPreferencesTool()
			cloud = prefs.get('private', 'upstreambackup')
			files = []
			if tableobj or hasattr(self, tablename):
				view = tableobj
				if not(tableobj):
					view = getattr(self, tablename)
				props = view.structure()
				cols = []
				for prop in props:
					if not(excols.has_key(prop.name)):
						cols.append(prop.name)
				sortparms = []
				for col in sortcols:
					if col in cols:
						sortparms.append(getattr(view, col))
				view = apply(view.sort,sortparms)
				out = StringIO.StringIO()
				last = None
				lastcols = []
				for row in view:
					currentcols = []
					valuelist = []
					for el in sortcols:
						valuelist.append(getattr(row, el))
						currentcols.append((el, getattr(row, el)))
					if selectfn:
						if not(selectfn(row)):
							break
					if groupfn:
						currentcols = groupfn(currentcols)
					current = tablename
					for el in currentcols:
						current += '-%s' % el[1]
					if last != current:
						out = out.getvalue()
						if out:
							files.append(_storeBackup(out, last, lastcols, cloud))
						out = StringIO.StringIO()
						last = current
						lastcols = currentcols
					out.write(' <item>\n')
					for col in cols:
						val = getattr(row, col)
						typ = _valType(val)
						if typ == 'PyView':
							cl = []
							for c in val.structure():
								cl.append(c.name)
							for v in val:
								out.write('  <%s type="%s">\n' % (col, typ))
								for c in cl:
									cv = getattr(v, c)
									out.write('   <%s type="%s">%s</%s>\n' % (c, _valType(cv), macros.quoteXML(cv, controlchars=1), c))
								out.write('  </%s>\n' % col)
						else:
							out.write('  <%s type="%s">%s</%s>\n' % (col, typ, macros.quoteXML(str(val), controlchars=1), col))
					out.write(' </item>\n')
				out = out.getvalue()
				if out:
					files.append(_storeBackup(out, last, lastcols, cloud))
			return files
		finally: self._release()

	def _startthread(self, status):

		def thunk(tool, server):
			failcount = 0
			lastfail = time.time()
			tool.logVerbose(_('Background process started'))
			while server.isRunning() and failcount < 5:
				try:
					try:
						_flet.beginCloud(current=self)
						tool._initthread(server)
					finally:
						_flet.end()
				except:
					(e, d, tb) = sys.exc_info()
					msg = _('Background thread exception %s: %s') % (e, d)
					for row in traceback.extract_tb(tb):
						(file, line, function, source) = row
						msg += '<br>%s[%s] in %s' % (file, line, function)
					tool.logError(msg.replace('%', '%%'))
					if lastfail < time.time() - 600:
						failcount = 0
					else:
						failcount += 1
					lastfail = time.time()
			tool.logVerbose(_('Background process ended'))
		
		background = threading.Thread(
			target=thunk,
			args=[self, status],
			kwargs={}
		)
		self._mythread = background
		self._serverstatus = status
		status.append(background)
		background.start()
		return background

	def _threadstatus(self):
		html = ''
		if self._mythread.isAlive():
			html += _('<li>The background thread is active</li>')
		else:
			html += _('<li>The background thread is <a href="%s" title="%s">inactive</a>') % (
				self.getUrl('startthread_redir'),
				_('restart the thread')
			)
		return html
	
	def startthread_redir(self, req):
		if hasattr(self, '_mythread'):
			if not(self._mythread.isAlive()):
				self._startthread(self._serverstatus)
		return req.getUrl(tool='status')

	def _initdb(self):
		pass
		
	def _initopts(self):
		self.rss = 0

	def _commit(self):
		self.db.commit()

	def _acquire(self, blocking=1):
		self.lock.acquire(blocking)
	
	def _release(self):
		self.lock.release()
	
	def _getDummyRequest(self, now=None, datespec=None, base=None, flat=0):
		params = None
		dateGiven = 0
		dayGiven = 0
		if not(now):
			now = time.time()
		else:
			params = '%04d-%02d-%02d' % time.localtime(now)[:3]
		if datespec:
			t = list(time.gmtime(now))
			for i in range(0,3):
				if datespec[i]:
					t[i] = datespec[i]
			now = time.mktime(t)
			if datespec[0] and datespec[1] and datespec[2]:
				params = '%04d-%02d-%02d' % time.localtime(now)[:3]
				dateGiven = 1
				dayGiven = 1
			elif datespec[0] and datespec[1]:
				params = '%04d-%02d' % time.localtime(now)[:2]
				dateGiven = 1
		req = _Request(self, now=now, base=base, flat=flat)
		if params:
			req.params = params
			req.dateGiven = dateGiven
			req.dayGiven = dayGiven
		return req
	
	def _getDirectory(self, **args):
		p = _PyDS.HOMEDIR
		for a in args:
			if type(a) == type(u''):
				a = a.encode('iso-8859-1')
			if not(a in ('.', '..')):
				p = os.path.join(p, a)
		os.makedirs(p)
		return p

	def _getLockObject(self):
		return threading.RLock()

	def _hookIntoRenderTool(self, indexf):
		try:
			self._acquire()
			render = self.getToolByNamespace('render')
			render._hookTool(self, indexf)
		finally: self._release()
	
	def _addHook(self, mode, tag, title, func, activeFunc=None):
		try:
			self._acquire()
			if self.hooks.has_key(mode):
				self.hooks[mode].append(StandardHook(self.name, mode, tag, title, func, activeFunc))
			else:
				self.hooks[mode] = [StandardHook(self.name, mode, tag, title, func, activeFunc)]
		finally: self._release()
	
	def _getHookModes(self):
		try:
			self._acquire()
			modes = self.hooks.keys()
			modes.sort()
			return modes
		finally: self._release()
	
	def _getHooks(self, mode):
		try:
			self._acquire()
			hooks = self.hooks.get(mode, [])
			hooks.sort(lambda a,b: cmp(a.title, b.title))
			return hooks
		finally: self._release()
	
	def _getHook(self, mode, name):
		try:
			self._acquire()
			for hook in self._getHooks(mode):
				if hook.getId() == name:
					return hook
			return None
		finally: self._release()
		
	def _status(self):
		if self.level < 2000:
			html = _('<p><b>"%s" is installed at level %d.</b></p>') % (
				self.name,
				self.level
			)
		else:
			html = _('<p><b>"%s" is not installed.</b></p>') % (
				self.name
			)
		return html
	
	def match(self, request):
		# <path>;<params>?<query>#<fragment>
		[path, params, query, fragment] = request.split_uri()
		name = '/' + self.name
		if self.level==0 and (path == '/' or path == ''):
			return 1
		elif (path == name) or (path == name + '/'):
			return 1
		else:
			return (self.regex.match(path) != None)

	def handle_request(self, request):
		if request.command.lower() in ('put', 'post'):
			cl = request.get_header('content-length')
			length = int(cl)
			if not cl:
				request.error(411)
			else:
				Collector(self, length, request)
		else:
			self.continue_request(request, StringIO.StringIO())

	def continue_request(self, request, input_data):
		req = _Request(self, request=request, input_data=input_data)
		user = self.getToolByNamespace('user')
		if hasattr(request, 'authenticated_user'):
			user.authenticated_user = request.authenticated_user
		else:
			user.authenticated_user = ''
		name = '/' + self.name
		if self.name != 'preferences' and request.command.lower() != 'post':
			stuff = _ToolNamespace['preferences'].needsStuff(req)
			if stuff:
				request['Location'] = stuff
				request.error(302)
				return
			for toolname in _ToolNamespace.keys():
				if toolname != 'preferences':
					tool = _ToolNamespace[toolname]
					stuff = tool.needsStuff(req)
					if  stuff:
						request['Location'] = stuff
						request.error(302)
						return
		if self.level==0 and (req.path == '/' or req.path == ''):
			try:
				try:
					_flet.beginDesktop(current=self)
					s = self.index_html(req)
				finally:
					_flet.end()
				request['Content-type'] = 'text/html'
			except:
				e = sys.exc_info()
				return self.error(request, e)
		elif (req.path == name) or (req.path == name + '/'):
			try:
				try:
					_flet.beginDesktop(current=self)
					s = self.index_html(req)
				finally:
					_flet.end()
				request['Content-type'] = 'text/html'
			except:
				e = sys.exc_info()
				return self.error(request, e)
		else:
			match = self.regex.match(req.path)
			if match:
				meth = match.group(1)
				if meth[0] == '_':
					request.error(404)
					return
				elif hasattr(self,meth):
					if meth[-5:] == '_html':
						request['Content-type'] = 'text/html'
						try:
							try:
								_flet.beginDesktop(current=self)
								s = getattr(self,meth)(req)
							finally:
								_flet.end()
						except:
							e = sys.exc_info()
							return self.error(request, e)
					elif meth[-4:] == '_xml':
						request['Content-type'] = 'text/xml'
						try:
							_flet.beginDesktop(current=self)
							s = getattr(self,meth)(req)
						finally:
							_flet.end()
					elif meth[-4:] == '_jpg':
						request['Content-type'] = 'image/jpg'
						try:
							_flet.beginDesktop(current=self)
							s = getattr(self,meth)(req)
						finally:
							_flet.end()
					elif meth[-4:] == '_png':
						request['Content-type'] = 'image/png'
						try:
							_flet.beginDesktop(current=self)
							s = getattr(self,meth)(req)
						finally:
							_flet.end()
					elif meth[-6:] == '_redir':
						try:
							try:
								_flet.beginCloud(current=self)
								u = getattr(self,meth)(req)
							finally:
								_flet.end()
						except:
							e = sys.exc_info()
							return self.error(request, e)
						s = '<html><head><meta http-equiv="Refresh" content="0;url=%s"></head></html>' % u
						request['Content-type'] = 'text/html'
					else:
						request.error(404)
						return
				else:
					request.error(404)
					return
			else:
				request.error(404)
				return
		request['Content-length'] = len(s)
		request.push(s)
		request.done()

	def error(self, request, exc_info, format='html'):
		(e, d, tb) = exc_info
		if format == 'html':
			if request:
				request['Content-type'] = 'text/html'
			s = ''
			if request:
				s += _('<html><head><title>Exception %s: %s</title></head>') % (e, d)
				s += '<body>'
				s += _('<h1>Exception %s: %s</h1>') % (e, d)
			else:
				s += _('<b>Exception %s: %s</b>') % (e, d)
			s += '<table border=0>'
			s += _('<tr><th>File</th><th>Line</th><th>Function</th><th>Source</th></tr>')
			for row in traceback.extract_tb(tb):
				s += '<tr>'
				(file, line, function, source) = row
				file = file.split('/')[-1]
				s += '<td>%s</td>' % file
				s += '<td>%s</td>' % line
				s += '<td>%s</td>' % function
				s += '<td>%s</td>' % source
				s += '</tr>'
			s += '</table>'
			if request:
				s += '</body></html>'
				request['Content-length'] = len(s)
				request.push(s)
				request.done()
			else:
				return s
		elif format == 'xml':
			request['Content-type'] = 'text/xml'
		else:
			raise ValueError('Unknown error format: %s' % format)

	# this method searches for a string, if the tool supports searching
	def search(self, query):
		if hasattr(self, '_search'):
			return self._search(query)
		else:
			return []
	
	# this method renders search results from sequential searches
	# on tools databases
	def searchresults_html(self, req):
		html = '<table border=0>'
		for item in self._search(req.getFirstValue('query')):
			guid = ''
			if item.get('guid'):
				guid = '<a href="%s" target="_blank" title="%s"><img src="/images/remote.gif" border=0></a> ' % (
					item['guid'],
					_('Click here to go to the cloud page')
				)
			html += '<tr><td>%s<a href="%s">%s</a></td></tr>' % (
				guid,
				item['link'],
				item['title'] or item['link'],
			)
			if item.get('description'):
				html += '<tr><td class="smallfont">%s</td></tr>' % item['description']
		html += '</table>'
		req.setLocalValue('title', _('Your search results'))
		return req.renderPage('BaseTemplate', body=html)

	# This method tells the server that this tool needs additional
	# data. This can be overridden by subclasses to hook up mandatory
	# preferences
	def needsStuff(self, req):
		return 0
			
	def index_html(self, req):
		html = ''
		for tool in getTools():
			html += '<li>%s</li>' % tool.title
		html += "</ul>"
		req.setLocalValue('title', _('Tool Index'))
		return req.renderPage('BaseTemplate', body=html)
	
	# this method returns a searchbox if the tool supports searching
	def searchbox(self):
		if hasattr(self, '_search'):
			html = '<form class="whiteboxsmall" action="%s" method="GET">' % self.getUrl('searchresults_html')
			html += '<input name="query" size=10>'
			html += ' <input type="image" name="search" src="/images/search.gif">'
			html += '</form>'
			return html
		else:
			return ''

	# this method returns a short description of a tool. Tools can
	# override this method to give more descriptive text.
	def nugget(self, url=None):
		if not(url):
			url = self.url
		html = '<a href="%s">%s</a>: %s' % (url, self.title, self.desc)
		return html

	# this method logs an event to the registered events tool. You need
	# to give a string format and optional arguments. The message is
	# automatically timestamped and marked with the current tool
	def log(self, fmt, *args):
		msg = fmt % args
		_ToolNamespace['events']._log(self.name, msg)
		return 1

	# this method logs an error to the registered events tool. You need
	# to give a string format and optional arguments. The message is
	# automatically timestamped and marked with the current tool
	def logError(self, fmt, *args):
		msg = fmt % args
		_ToolNamespace['events']._log(self.name, msg, error=1)
		return 1

	# this method logs an event to the registered events tool. You need
	# to give a string format and optional arguments. The message is
	# automatically timestamped and marked with the current tool.
	# This message is only logged when verbosity is activated!
	def logVerbose(self, fmt, *args):
		msg = fmt % args
		_ToolNamespace['events']._logVerbose(self.name, msg)
		return 1

	# this methods returns the text and link that should be used in
	# the top menubar.
	def menuitem(self, active=1):
		if active:
			html = '<a href="%s" title="%s">%s</a>' % (self.url, self.desc, self.title)
		else:
			html = '%s' % (self.title, )
		return html
		
	# this method parses a XML table and returns the tables name
	# and it's items
	def parseBackup(self, url):
		f = None
		if len(url) > 7 and url[:7] == 'http://':
			downstream = self.getToolByNamespace('downstream')
			f = downstream.download(url)
		else:
			f = open(url)
		content = f.read()
		f.close()
		p = BackupParser()
		p.feed(content)
		return [p._toolname, p._tablename, p._items]
	
	# this method returns the title for BaseRendering.tmpl
	# this can be overridden in the tool if there are special requirements.
	# the context should be set to a dictionary with additional parameters
	# (might be $context in renderPage calls)
	def getPageTitle(self, context):
		prefs = self.getPreferencesTool()
		title = prefs.get('private', 'title')
		return title
		
	# this method returns the description for BaseRendering.tmpl
	# this can be overridden in the tool if there are special requirements.
	# the context should be set to a dictionary with additional parameters
	# (might be $context in renderPage calls)
	def getPageDescription(self, context):
		prefs = self.getPreferencesTool()
		desc = prefs.get('private', 'description')
		return desc
	
	# get all dates where you have items - usually overridden by a tool. This is
	# used by the calendar tool to cache day data. Category can be None for all
	# items, '' for the default category (onhome=1 in the weblog) or a category name.
	# Must return a tuple of year, month and day for each day that has items.
	def getItemDays(self, category):
		return []
		
	# check whether a day has any items at all for a given category. This is for
	# updates that initialize a recaching
	def hasItemsForDay(self, year, month, day, category):
		return 0
	
	# gives a filename for a file within a tool
	def getFileName(self, name, category='', toolname=None, cloud=1):
		if toolname == None:
			toolname = self.name
		if cloud:
			return os.path.join(_PyDS.ROOTDIR, toolname, category, name)
		else:
			return os.path.join(_PyDS.HOMEDIR, toolname, category, name)
		
	# this method stores a given HTML source in the specified file.
	# It uses the date specifier to construct a filename.
	# Upstreaming is triggered.
	def storeHTML(self, html, page='index.html', year=None, month=None, day=None, category='', toolname=None):
		mainindex = 0
		if toolname == None:
			toolname = self.name
		if year and month and day:
			dn = os.path.join(_PyDS.ROOTDIR, toolname, category,
				'%04d' % year,
				'%02d' % month)
			fn = os.path.join(dn, '%02d.html' % day)
		elif year and month:
			dn = os.path.join(_PyDS.ROOTDIR, toolname, category,
				'%04d' % year,
				'%02d' % month)
			fn = os.path.join(dn, page)
		else:
			dn = os.path.join(_PyDS.ROOTDIR, toolname, category)
			fn = os.path.join(dn, page)
			if not(category):
				mainindex = 1
		if dn[-1] == '/':
			dn = dn[:-1]
		if not(os.path.exists(dn)):
			os.makedirs(dn)
		old = ''
		if os.path.exists(fn):
			f = open(fn)
			old = f.read()
			f.close()
		if old != html:
			f = open(fn, 'w')
			f.write(html)
			f.close()
			self.log(_('written file: %s') % fn[len(_PyDS.ROOTDIR):])
			_ToolNamespace['upstream'].awakeUpstream()
		if self.level == 0 and mainindex and page == 'index.html':
			old = ''
			fn = os.path.join(_PyDS.ROOTDIR, page)
			if os.path.exists(fn):
				f = open(fn)
				old = f.read()
				f.close()
			if old != html:
				f = open(fn, 'w')
				f.write(html)
				f.close()
				self.log(_('written file: %s') % fn[len(_PyDS.ROOTDIR):])
				_ToolNamespace['upstream'].awakeUpstream()
		return fn
				
	# this method stores a given XML source in the specified file. cloud=0
	# would tell the system to store in a local (not upstreamed) area.
	# Upstreaming is triggered if the file should go to the cloud.
	def storeXML(self, xml, name='rss.xml', category='', toolname=None, cloud=1):
		xmlfn = self.storeFile(xml, name, category, toolname, cloud=cloud)
		return xmlfn

	# this method stores a given file content in the specified file. cloud=0
	# would tell the system to store in a local (not upstreamed) area.
	# Upstreaming is triggered if the file should go to the cloud.
	def storeFile(self, content, name, category='', toolname=None, cloud=1):
		if toolname == None:
			toolname = self.name
		fn = self.getFileName(name, category, toolname, cloud=cloud)
		old = ''
		if os.path.exists(fn):
			f = open(fn)
			old = f.read()
			f.close()
		if content != old:
			dn = os.path.dirname(fn)
			if not(os.path.isdir(dn)):
				os.makedirs(dn)
			f = open(fn, 'wb')
			f.write(content)
			f.close()
			if cloud:
				self.log(_('written file: %s') % fn[len(_PyDS.ROOTDIR):])
				_ToolNamespace['upstream'].awakeUpstream()
			else:
				self.log(_('written local file: %s') % fn[len(_PyDS.HOMEDIR):])
		return fn
			
	# this method deletes a file in the upstream directory
	def deleteFile(self, name, category='', toolname=None):
		if not(toolname):
			toolname = self.name
		fn = os.path.join(_PyDS.ROOTDIR, toolname, name)
		if category:
			fn = os.path.join(_PyDS.ROOTDIR, toolname, category, name)
		old = ''
		if os.path.exists(fn):
			os.unlink(fn)
			self.log(_('deleted file: %s') % fn[len(_PyDS.ROOTDIR):])
			_ToolNamespace['upstream'].awakeUpstream()
		return fn

	# this method returns an URL in the context of the current tool
	# or the given tool. You can pass in additional parameters that are
	# added to the querystring.
	def getUrl(self, page='index_html', tool=None, fragment=None, **kw):
		if tool:
			tool = self.getToolByNamespace(tool)
		else:
			tool = self
		url = '/%s/%s' % (tool.name, page)
		l = []
		for k in kw.keys():
			kv = kw[k]
			if type(kv) != type(''):
				kv = str(kv)
			kv = urllib.quote_plus(kv)
			l.append('%s=%s' % (k, kv))
		if len(l):
			url += '?' + string.join(l, '&')
		if fragment:
			url += '#%s' % fragment
		return url
				
	# this method returns an URL in the context of the current tool
	# or the given tool. You can pass in additional parameters that are
	# added to the querystring. This is for rendered output. You can
	# provide a base URL if you need to. If you provide flat=true, you
	# get flat filenames instead of structured directories. This is
	# usefull for exports to CD-ROM or stuff like that. In that case,
	# base is usually '/'
	def getStaticUrl(self, page=None, tool=None, year=None, month=None, day=None, category='', base=None, flat=0):
		if not(page) and year and month and day:
			page = '%04d/%02d/%02d.html' % (year, month, day)
		elif not(page) and year and month:
			page = '%04d/%02d/index.html' % (year, month)
		elif not(page):
			page = 'index.html'
		if tool:
			tool = self.getToolByNamespace(tool)
		else:
			tool = self
		if not(base):
			upstream = self.getToolByNamespace('upstream')
			base = upstream.get('prefs','cloudurl')
			if not(base):
				base = '/'
			if base[-1] != '/':
				base += '/'
		if flat:
			if category:
				return '%s%s-%s-%s' % (base, tool.name, category, page)
			else:
				return '%s%s-%s' % (base, tool.name, page)
		else:
			url = ''
			if category:
				url = '%s%s/%s/%s' % (base, tool.name, category, page)
			else:
				url = '%s%s/%s' % (base, tool.name, page)
			if len(url) >= 11 and url[-11:] == '/index.html':
				url = url[:-10]
			return url
		
	# this returns the installed preferences tool
	def getPreferencesTool(self):
		return _ToolNamespace['preferences']
		
	# this returns the tool that installs into the given namespace
	def getToolByNamespace(self, namespace):
		return _ToolNamespace.get(namespace, None)
	
	# this method returns a list with method names without _ in them
	def listRegisteredFunctions(self):
		liste = []
		for el in dir(self):
			if el.find('_') < 0:
				if hasattr(getattr(self, el), '__call__'):
					liste.append(el)
		return liste
	
	# this method usually returns a list of hours without RSS items.
	# you want to overload it, if your tool want's to produce skiphours
	# elements in the RSS feed.
	def getRSSSkipHours(self, category=''):
		return []

	# this method uses getRSSItems to find the newest date that
	# is given in the items. This is used for lastBuildDate. Usually
	# this implementation is sufficient and doesn't need to be
	# overloaded. If there are no items, the date returned is the
	# empty string! This method doesn't need sorted items, so it will
	# work for rss feeds with different sort order (not newest item at
	# the top), too.
	def getRSSNewestDate(self, category=''):
		macros = self.getToolByNamespace('macros')
		maxdate = 0
		for item in self.getRSSItems(category=category):
			date = macros.parseTodayFull(item['pubdate'])
			if date > maxdate:
				maxdate = date
		if maxdate:
			return macros.TodayFull(maxdate)
		else:
			return ''

