"""

Python Desktop Server - Topic Exchange integration

Copyright (c) 2002, Georg Bauer <gb@murphy.bofh.ms>

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 
the Software, and to permit persons to whom the Software is furnished to do so, 
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""

# $Id: TopicExchangeTool.py,v 1.19 2003/06/12 12:21:17 gb Exp $

import time

import PyDS.Tool

class TopicExchangeTool(PyDS.Tool.StandardTool):

	def _initdb(self):
		try:
			self._acquire()
			self.channels = self.db.getas(
				"topicchannels[topic:S,url:S]"
			).ordered(1)
			self._commit()
		finally: self._release()
	
	def _backup(self):
		return self._backupTable('channels')
	
	def _initopts(self):
		upstream = self.getToolByNamespace('upstream')
		self.server = upstream.getServerForXMLRPC('http://topicexchange.com/RPC2')
		weblog = self.getToolByNamespace('weblog')
		weblog._addHook('edit', 'select', _('Topics'), self._topicSelection, self._topicSelectionActive)
		weblog._addHook('editold', 'ping', '', self._topicPingHook, self._topicSelectionActive)
		weblog._addHook('postnew', 'ping', '', self._topicPingHook, self._topicSelectionActive)

	def _initthread(self, server):
		while server.isRunning():
			time.sleep(10)
			if self.queue.length():
				while self.queue.length():
					(jobf, key, data) = self.queue.shift()
					jobf(key, data)
				self.logVerbose(_('work done, going to sleep again'))
	
	def _topicPingHook(self, tool, req, id, *args, **kw):
		try:
			self._acquire()
			topics = req.getAllValues('ptopic')
			for topic in topics:
				(idx, found) = self.channels.locate({'topic':topic})
				if found > 0:
					self.queue.append(
						self._topicPingTopic,
						(tool.name, id, self.channels[idx].topic),
						tool.getRSSItem(id)
					)
		finally: self._release()
	
	def _topicPingTopic(self, key, data):
		try:
			self._acquire()
			prefs = self.getPreferencesTool()
			macros = self.getToolByNamespace('macros')
			(toolname, id, topic) = key
			item = data
			if item:
				excerpt = macros.unquoteXML(item['description'])
				excerpt = macros.stripHTML(excerpt)
				excerpt = macros.quoteXML(excerpt)
				if len(excerpt) > 250:
					excerpt = excerpt[:250] + ' ...'
				details = {
					'blog_name': prefs.get('private','title'),
					'title': item['title'],
					'url': item['guid'],
					'excerpt': excerpt,
					'autoMunge': 0,
				}
				res = self.server.topicExchange.ping(topic, details)
				if res['flError']:
					self.log(_('topicExchange.ping gave error %s'), res['message'])
				else:
					self.log(_('topicExchange.ping says %s'), res['message'])
					self.log(_('topic <a href="%s" target="_blank">%s</a> changed'), res['topicUrl'], res['topicName'])
			else:
				self.log(_('trackback ping with null object for topic %s with item %s from tool %s'), topic, id, toolname)
		finally: self._release()
	
	def _topicSelection(self, *args, **kw):
		html = ''
		for channel in self.getActiveChannels():
			html += '<input type=checkbox name="ptopic" value="%s"> %s' % (channel['topic'], channel['topic'])
		return html
	
	def _topicSelectionActive(self, *args, **kw):
		try:
			self._acquire()
			return len(self.channels)
		finally: self._release()
	
	def _status(self):
		try:
			self._acquire()
			html = PyDS.Tool.StandardTool._status(self)
			html += '<ul>'
			html += _('<li>There are %d pings in the queue</li>') % self.queue.length()
			html += self._threadstatus()
			html += '</ul>'
			return html
		finally: self._release()
	
	# this method returns all channels at topicexchange.com
	def getChannels(self):
		channels = []
		res = self.server.topicExchange.getChannels()
		if res['flError']:
			self.log(_('topicExchange.getChannels gave %s'), res['message'])
		else:
			cdict = res['channels']
			for k in cdict.keys():
				channels.append({
					'topic': k,
					'url': cdict[k]['url']
				})
			channels.sort(lambda a,b: cmp(a['topic'], b['topic']))
		return channels
	
	# this method checks wether we selected a channel as active
	def isChannelActive(self, topic):
		try:
			self._acquire()
			(idx, found) = self.channels.locate({'topic':topic})
			return found > 0
		finally: self._release()

	# this method returns a list of active channels
	def getActiveChannels(self):
		try:
			self._acquire()
			channels = []
			for channel in self.channels:
				channels.append({
					'topic': channel.topic,
					'url': channel.url
				})
			channels.sort(lambda a,b: cmp(a['topic'],b['topic']))
			return channels
		finally: self._release()

	def getPreferencePanels(self):
		return [
			['topic_panel', _('Topics'),
				_('Here you can preselect TopicExchange topics for postings.')]
		]
			
	def topic_panel(self, req):
		macros = self.getToolByNamespace('macros')
		html = _('<p>All topics at topicexchange.com')
		html += '<form action="%s" method="POST">' % self.getUrl('activate_redir')
		html += '<table border=0 width="100%">'
		html += '<tr>'
		html += '<th>%s</th><th align="left">%s</th>' % (_('Subscribe'), _('Title'))
		html += '</tr>'
		for channel in self.getChannels():
			html += '<tr>'
			html += '<td>'
			if self.isChannelActive(channel['topic']):
				html += '<input type=checkbox name="channel" value="%s" checked>' % channel['topic']
			else:
				html += '<input type=checkbox name="channel" value="%s">' % channel['topic']
			html += '<input type=hidden name="ctopic" value="%s">' % channel['topic']
			html += '<input type=hidden name="curl_%s" value="%s">' % (channel['topic'], channel['url'])
			html += '</td>'
			html += '<td><a target="_blank" href="%s">%s</a></td>' % (channel['url'], channel['topic'])
			html += '</tr>'
		html += '</table>'
		html += '<input type=submit name="TopicExchangeSubmit" value="%s">' % _('save')
		html += '</form>'
		return html
	
	def activate_redir(self, req):
		try:
			self._acquire()
			curls = {}
			cactive = {}
			for topic in req.getAllValues('ctopic'):
				curls[topic] = req.getFirstValue('curl_%s' % topic)
				cactive[topic] = 0
			for topic in req.getAllValues('channel'):
				cactive[topic] = 1
				(idx, found) = self.channels.locate({'topic':topic})
				if found > 0:
					self.channels[idx].url = curls[topic]
				else:
					self.channels.append({
						'topic': topic,
						'url': curls[topic]
					})
			todel = []
			for idx in range(len(self.channels)):
				channel = self.channels[idx]
				if not(cactive.get(channel.topic, 0)):
					todel.append(idx)
			todel.sort(lambda a,b: -1*cmp(a,b))
			for idx in todel:
				self.channels.delete(idx)
			self._commit()
			return req.getUrl(
				'panel_html',
				tool='preferences',
				panel='topic_panel',
				fortool='topicexchange'
			)
		finally: self._release()

__desc__ = _("This tool manages topics from topicexchange.com.")

PyDS.Tool.registerTool(TopicExchangeTool('topicexchange', _('TopicExchange'), __desc__, 1300))

